///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//  Copyright  NetworkDLS 2002, All rights reserved
//
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF 
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO 
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A 
// PARTICULAR PURPOSE.
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#ifndef _CLinkedList_CPP_
#define _CLinkedList_CPP_
////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <Windows.H>
#include <STDIO.H>
#include <STDLib.H>

////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "CLinkedList.H"

////////////////////////////////////////////////////////////////////////////////////////////////////////////

CLinkedList::CLinkedList(void)
{
	this->NodeCount = 0;
	this->RootNode = NULL;
	this->LastNode = NULL;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

CLinkedList::~CLinkedList(void)
{
	this->DeleteAllNodes();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CLinkedList::MoveNodeUp(STACKNODE *pNode)
{
	STACKNODE *Parent = pNode->Parent;
	if(Parent)
	{
		unsigned int pOldParent = (int)Parent->Parent;
		unsigned int pOldChild  = (int)pNode->Child;

		if(Parent->Parent)
		{
			Parent->Parent->Child = pNode;
		}
		if(pNode->Child)
		{
			pNode->Child->Parent = Parent;
		}

		pNode->Child = Parent;
		Parent->Parent = pNode;

		pNode->Parent = (STACKNODE *)pOldParent;
		Parent->Child = (STACKNODE *)pOldChild;

		if(!Parent->Child)
		{
			this->LastNode = Parent;
		}
		if(!Parent->Parent)
		{
			this->RootNode = Parent;
		}
		if(!pNode->Child)
		{
			this->LastNode = pNode;
		}
		if(!pNode->Parent)
		{
			this->RootNode = pNode;
		}
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CLinkedList::SortAsc(void)
{
	while(!this->SortAsc(this->NodeCount))
	{
		//Keep on sorting.
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CLinkedList::SortAsc(unsigned int iMaxMoves)
{
	bool bMoved = true;
	bool bMove = false;
	unsigned int iMoves = 0;

	while(bMoved)
	{
		bMoved = !bMoved;

		STACKNODE *pNode = this->RootNode;
		while(pNode)
		{
			if(pNode->Parent)
			{
				bMove = false;

				if(pNode->Data.Type == NODE_TYPE_TEXT)
					bMove = (strcmp((char *)pNode->Data.Buffer, (char *)pNode->Parent->Data.Buffer) < 0);
				if(pNode->Data.Type == NODE_TYPE_WTEXT)
					bMove = (wcscmp((wchar_t *)pNode->Data.Buffer, (wchar_t *)pNode->Parent->Data.Buffer) < 0);
				else if(pNode->Data.Type == NODE_TYPE_SINT16)
					bMove = (*((signed short *)pNode->Data.Buffer) < *((signed short *)pNode->Parent->Data.Buffer));
				else if(pNode->Data.Type == NODE_TYPE_UINT16)
					bMove = (*((unsigned short *)pNode->Data.Buffer) < *((unsigned short *)pNode->Parent->Data.Buffer));
				else if(pNode->Data.Type == NODE_TYPE_SINT32)
					bMove = (*((signed int *)pNode->Data.Buffer) < *((signed int *)pNode->Parent->Data.Buffer));
				else if(pNode->Data.Type == NODE_TYPE_UINT32)
					bMove = (*((unsigned int *)pNode->Data.Buffer) < *((unsigned int *)pNode->Parent->Data.Buffer));
				else if(pNode->Data.Type == NODE_TYPE_SINT64)
					bMove = (*((signed __int64 *)pNode->Data.Buffer) < *((signed __int64 *)pNode->Parent->Data.Buffer));
				else if(pNode->Data.Type == NODE_TYPE_UINT64)
					bMove = (*((unsigned __int64 *)pNode->Data.Buffer) < *((unsigned __int64 *)pNode->Parent->Data.Buffer));
				else if(pNode->Data.Type == NODE_TYPE_FLOAT)
					bMove = (*((float *)pNode->Data.Buffer) < *((float *)pNode->Parent->Data.Buffer));
				else if(pNode->Data.Type == NODE_TYPE_DOUBLE)
					bMove = (*((double *)pNode->Data.Buffer) < *((double *)pNode->Parent->Data.Buffer));
				else if(pNode->Data.Type == NODE_TYPE_SBYTE)
					bMove = (*((signed char *)pNode->Data.Buffer) < *((signed char *)pNode->Parent->Data.Buffer));
				else if(pNode->Data.Type == NODE_TYPE_UBYTE)
					bMove = (*((unsigned char *)pNode->Data.Buffer) < *((unsigned char *)pNode->Parent->Data.Buffer));

				if(bMove)
				{
					this->MoveNodeUp(pNode);
					bMoved = true;
					iMoves++;
				}
			}
			pNode = pNode->Child;
		}
		if(iMoves > iMaxMoves)
		{
			return false;
		}
	}
	return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CLinkedList::SortDes(void)
{
	while(!this->SortDes(this->NodeCount))
	{
		//Keep on sorting.
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CLinkedList::SortDes(unsigned int iMaxMoves)
{
	bool bMoved = true;
	bool bMove = false;
	unsigned int iMoves = 0;

	while(bMoved)
	{
		bMoved = !bMoved;

		STACKNODE *pNode = this->RootNode;
		while(pNode)
		{
			if(pNode->Parent)
			{
				bMove = false;

				if(pNode->Data.Type == NODE_TYPE_TEXT)
					bMove = (strcmp((char *)pNode->Data.Buffer, (char *)pNode->Parent->Data.Buffer) > 0);
				if(pNode->Data.Type == NODE_TYPE_WTEXT)
					bMove = (wcscmp((wchar_t *)pNode->Data.Buffer, (wchar_t *)pNode->Parent->Data.Buffer) > 0);
				else if(pNode->Data.Type == NODE_TYPE_SINT16)
					bMove = (*((signed short *)pNode->Data.Buffer) > *((signed short *)pNode->Parent->Data.Buffer));
				else if(pNode->Data.Type == NODE_TYPE_UINT16)
					bMove = (*((unsigned short *)pNode->Data.Buffer) > *((unsigned short *)pNode->Parent->Data.Buffer));
				else if(pNode->Data.Type == NODE_TYPE_SINT32)
					bMove = (*((signed int *)pNode->Data.Buffer) > *((signed int *)pNode->Parent->Data.Buffer));
				else if(pNode->Data.Type == NODE_TYPE_UINT32)
					bMove = (*((unsigned int *)pNode->Data.Buffer) > *((unsigned int *)pNode->Parent->Data.Buffer));
				else if(pNode->Data.Type == NODE_TYPE_SINT64)
					bMove = (*((signed __int64 *)pNode->Data.Buffer) > *((signed __int64 *)pNode->Parent->Data.Buffer));
				else if(pNode->Data.Type == NODE_TYPE_UINT64)
					bMove = (*((unsigned __int64 *)pNode->Data.Buffer) > *((unsigned __int64 *)pNode->Parent->Data.Buffer));
				else if(pNode->Data.Type == NODE_TYPE_FLOAT)
					bMove = (*((float *)pNode->Data.Buffer) > *((float *)pNode->Parent->Data.Buffer));
				else if(pNode->Data.Type == NODE_TYPE_DOUBLE)
					bMove = (*((double *)pNode->Data.Buffer) > *((double *)pNode->Parent->Data.Buffer));
				else if(pNode->Data.Type == NODE_TYPE_SBYTE)
					bMove = (*((signed char *)pNode->Data.Buffer) > *((signed char *)pNode->Parent->Data.Buffer));
				else if(pNode->Data.Type == NODE_TYPE_UBYTE)
					bMove = (*((unsigned char *)pNode->Data.Buffer) > *((unsigned char *)pNode->Parent->Data.Buffer));

				if(bMove)
				{
					this->MoveNodeUp(pNode);
					bMoved = true;
					iMoves++;
				}
			}
			pNode = pNode->Child;
		}
		if(iMoves > iMaxMoves)
		{
			return false;
		}
	}
	return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

STACKNODE *CLinkedList::Next(STACKNODE *pNode)
{
	return pNode->Child;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

STACKNODE *CLinkedList::Prev(STACKNODE *pNode)
{
	return pNode->Parent;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CLinkedList::DestroyNode(STACKNODE *pNode)
{
	if(pNode->Data.Buffer)
	{
		free(pNode->Data.Buffer);
	}

	free(pNode);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CLinkedList::DeleteAllNodes(void)
{
	while(this->RootNode)
	{
		this->DeleteNode(this->RootNode);
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CLinkedList::DeleteNode(STACKNODE *pNode)
{
	if(pNode->Child) //Does this node have children?
	{
		//Assign this node's child to this node's parent.
		pNode->Child->Parent = pNode->Parent;
	}
	else{
		this->LastNode = pNode->Parent;
	}

	if(pNode->Parent) //Does this node have a parent? (If not, it's the Root node)
	{
		//Assign this node's parent to this node's child.
		pNode->Parent->Child = pNode->Child;
	}
	else{
		//The root node is going to be deleted, save it's child the Root.
		this->RootNode = pNode->Child;
	}

	this->DestroyNode(pNode);
	this->NodeCount--;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

STACKNODE *CLinkedList::CreateNode(const void *pData, unsigned int iDataLength, unsigned short iExtraAlloc)
{
	STACKNODE *pNode = (STACKNODE *) malloc(sizeof(STACKNODE));

	if(pNode)
	{
		pNode->Child = NULL;

		pNode->Data.Length = iDataLength;
		if((pNode->Data.Buffer = (char *) calloc(pNode->Data.Length + iExtraAlloc, 1)))
		{
			memcpy_s(pNode->Data.Buffer, pNode->Data.Length + iExtraAlloc, pData, pNode->Data.Length);
		}
		else { //Out of memory.
			#ifdef UNICODE
			this->Throw(L"Out of memory.");
			#else
			this->Throw("Out of memory.");
			#endif
			this->DeleteNode(pNode);
			return NULL;
		}

		if(this->NodeCount == 0)
		{
			this->RootNode = pNode;
		}
		else{
			this->LastNode->Child = pNode;
		}

		pNode->Parent = this->LastNode;

		this->LastNode = pNode;
		this->NodeCount++;
	}

	return pNode;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

STACKNODE *CLinkedList::Add(const void *pData, unsigned int iDataLength)
{
	return this->CreateNode(pData, iDataLength, 0);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

STACKNODE *CLinkedList::Add(const char *sText, unsigned int iTextLength)
{
	STACKNODE *pNode = this->CreateNode(sText, iTextLength, sizeof(char));
	pNode->Data.Type = NODE_TYPE_TEXT;
	return pNode;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

STACKNODE *CLinkedList::Add(const signed short is16)
{
	STACKNODE *pNode = this->Add(&is16, sizeof(is16));
	pNode->Data.Type = NODE_TYPE_SINT16;
	return pNode;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

STACKNODE *CLinkedList::Add(const unsigned short iu16)
{
	STACKNODE *pNode = this->Add(&iu16, sizeof(iu16));
	pNode->Data.Type = NODE_TYPE_UINT16;
	return pNode;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

STACKNODE *CLinkedList::Add(const signed int is32)
{
	STACKNODE *pNode = this->Add(&is32, sizeof(is32));
	pNode->Data.Type = NODE_TYPE_SINT32;
	return pNode;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

STACKNODE *CLinkedList::Add(const unsigned int iu32)
{
	STACKNODE *pNode = this->Add(&iu32, sizeof(iu32));
	pNode->Data.Type = NODE_TYPE_UINT32;
	return pNode;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

STACKNODE *CLinkedList::Add(const signed __int64 is64)
{
	STACKNODE *pNode = this->Add(&is64, sizeof(is64));
	pNode->Data.Type = NODE_TYPE_SINT64;
	return pNode;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

STACKNODE *CLinkedList::Add(const unsigned __int64 iu64)
{
	STACKNODE *pNode = this->Add(&iu64, sizeof(iu64));
	pNode->Data.Type = NODE_TYPE_UINT64;
	return pNode;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

STACKNODE *CLinkedList::Add(const float fNumber)
{
	STACKNODE *pNode = this->Add(&fNumber, sizeof(fNumber));
	pNode->Data.Type = NODE_TYPE_FLOAT;
	return pNode;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

STACKNODE *CLinkedList::Add(const double dNumber)
{
	STACKNODE *pNode = this->Add(&dNumber, sizeof(dNumber));
	pNode->Data.Type = NODE_TYPE_DOUBLE;
	return pNode;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

STACKNODE *CLinkedList::Add(const signed char isByte)
{
	STACKNODE *pNode = this->Add(&isByte, sizeof(isByte));
	pNode->Data.Type = NODE_TYPE_SBYTE;
	return pNode;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

STACKNODE *CLinkedList::Add(const unsigned char iuByte)
{
	STACKNODE *pNode = this->Add(&iuByte, sizeof(iuByte));
	pNode->Data.Type = NODE_TYPE_UBYTE;
	return pNode;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifdef UNICODE
void CLinkedList::Throw(const WCHAR *sText)
{
	MessageBox(GetActiveWindow(), sText, L"CLinkedList Exception", MB_ICONERROR);
	__debugbreak();
}
#else
void CLinkedList::Throw(const char *sText)
{
	MessageBox(GetActiveWindow(), sText, "CLinkedList Exception", MB_ICONERROR);
	__debugbreak();
}
#endif

////////////////////////////////////////////////////////////////////////////////////////////////////////////
#endif
