///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//  Copyright  NetworkDLS 2002, All rights reserved
//
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF 
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO 
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A 
// PARTICULAR PURPOSE.
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#ifndef _CLinkedList_H_
#define _CLinkedList_H_
////////////////////////////////////////////////////////////////////////////////////////////////////////////

#define NODE_TYPE_UNDEFINED	0
#define NODE_TYPE_TEXT		1
#define NODE_TYPE_WTEXT		2
#define NODE_TYPE_SINT16	3
#define NODE_TYPE_UINT16	4
#define NODE_TYPE_SINT32	5
#define NODE_TYPE_UINT32	6
#define NODE_TYPE_SINT64	7
#define NODE_TYPE_UINT64	8
#define NODE_TYPE_FLOAT		9
#define NODE_TYPE_DOUBLE	10
#define NODE_TYPE_SBYTE		11
#define NODE_TYPE_UBYTE		12

////////////////////////////////////////////////////////////////////////////////////////////////////////////

class CLinkedListNodeData {
public:
	void *Buffer;
	unsigned int Length;
	unsigned short Type;

	__inline signed short ToSInt16(void)
	{
		return *((signed short *)this->Buffer);
	}

	__inline unsigned short ToUInt16(void)
	{
		return *((unsigned short *)this->Buffer);
	}

	__inline signed int ToSInt32(void)
	{
		return *((signed int *)this->Buffer);
	}

	__inline unsigned int ToUInt32(void)
	{
		return *((unsigned int *)this->Buffer);
	}

	__inline signed __int64 ToSInt64(void)
	{
		return *((signed __int64 *)this->Buffer);
	}

	__inline unsigned __int64 ToUInt64(void)
	{
		return *((unsigned __int64 *)this->Buffer);
	}

	__inline float ToFloat(void)
	{
		return *((float *)this->Buffer);
	}

	__inline double ToDouble(void)
	{
		return *((double *)this->Buffer);
	}

	__inline signed char ToSByte(void)
	{
		return *((signed char *)this->Buffer);
	}

	__inline unsigned char ToUByte(void)
	{
		return *((unsigned char *)this->Buffer);
	}

	__inline char *ToString(void)
	{
		return ((char *)this->Buffer);
	}

	__inline wchar_t ToWChar(void)
	{
		return *((wchar_t *)this->Buffer);
	}

	__inline wchar_t *ToWString(void)
	{
		return ((wchar_t *)this->Buffer);
	}
};

struct STACKNODE {
	STACKNODE *Parent;
	STACKNODE *Child;
	CLinkedListNodeData Data;
};

////////////////////////////////////////////////////////////////////////////////////////////////////////////

class CLinkedList {

public:
	CLinkedList(void);
	~CLinkedList(void);

	STACKNODE *RootNode;
	STACKNODE *LastNode;

	int NodeCount;

	STACKNODE *Add(const void *pData, unsigned int iDataLength);
	STACKNODE *Add(const char *sText, unsigned int iTextLength);
	STACKNODE *Add(const signed short is16);
	STACKNODE *Add(const unsigned short iu16);
	STACKNODE *Add(const signed int is32);
	STACKNODE *Add(const unsigned int iu32);
	STACKNODE *Add(const signed __int64 is64);
	STACKNODE *Add(const unsigned __int64 iu64);
	STACKNODE *Add(const float fNumber);
	STACKNODE *Add(const double dNumber);
	STACKNODE *Add(const signed char isByte);
	STACKNODE *Add(const unsigned char iuByte);

	STACKNODE *Prev(STACKNODE *pNode);
	STACKNODE *Next(STACKNODE *pNode);

	void SortDes(void);
	bool SortDes(unsigned int iMaxMoves);

	void SortAsc(void);
	bool SortAsc(unsigned int iMaxMoves);

	void DeleteNode(STACKNODE *pNode);
	void DeleteAllNodes(void);
	void MoveNodeUp(STACKNODE *pNode);

private:
	STACKNODE *CreateNode(const void *pData, unsigned int iDataLength, unsigned short iExtraAlloc);
	void DestroyNode(STACKNODE *pNode);

	#ifdef UNICODE
	void Throw(const WCHAR *sText);
	#else
	void Throw(const char *sText);
	#endif

};

////////////////////////////////////////////////////////////////////////////////////////////////////////////
#endif
